;-----------------------------------------------------------------------------------------
; Test - DMA Fill FIFO Usage
;-Description:
; This test is designed to test what effect a DMA fill operation has on the FIFO. Hardware
; tests indicate that a DMA fill has no effect on the FIFO. The initial data port write
; which triggers the fill operation goes into the FIFO and gets written to VRAM like any
; normal write, after which, the DMA fill operation itself seems to have no effect on the
; VDP FIFO state.
;##NOTE## This test has been seen to occasionally lock up on the hardware, as well as
;produce different results on each run.
;-----------------------------------------------------------------------------------------
TestDMAFillFIFOUsageText:
	dc.b	'DMA Fill FIFO Usage                 '
TestDMAFillFIFOUsageTextEnd:
TestDMAFillFIFOUsageExpectedResult:
	dc.w $0000, $0000, $0000, $0000, $0000, $0000, $1000, $1000, $1234, $1212, $1212, $1212, $1212, $0012, $0000, $0000
TestDMAFillFIFOUsage:
	;Initialize the text result block
	move.w	#(TestDMAFillFIFOUsageTextEnd-TestDMAFillFIFOUsageText),D7
	move.w	#2,(A0)+ ;Set the number of display lines
	move.w	D7,(A0)+ ;Set the text length
	move.w	#$0020,D6 ;Set the data length
	move.w	D6,(A0)+
	lea		TestDMAFillFIFOUsageText,A1
	sub.w	#1,D7
	TestDMAFillFIFOUsageWriteText:
	move.b	(A1)+,(A0)+
	dbf	D7,TestDMAFillFIFOUsageWriteText
	move.w	#1,(A0)+ ;Set the expected result data present flag
	lea		TestDMAFillFIFOUsageExpectedResult,A1 ;Write the expected result data to the buffer
	sub.w	#1,D6
	TestDMAFillFIFOUsageWriteExpectedResultData:
	move.b	(A1)+,(A0)+
	dbf	D6,TestDMAFillFIFOUsageWriteExpectedResultData

	;Clear data in VRAM at 0x8000
	move.l	#$40000002,$C00004
	move.w	#$0000,$C00000
	move.w	#$0000,$C00000
	move.w	#$0000,$C00000
	move.w	#$0000,$C00000
	move.w	#$0000,$C00000
	move.w	#$0000,$C00000
	move.w	#$0000,$C00000
	move.w	#$0000,$C00000

	;Do a DMA fill operation to VRAM address 0x8000
	move.w #$000A,D1
	move.w #$8F01,$C00004
	move.w #$8154,$C00004
	move.w #$9300, D2
	move.b D1,D2
	move.w D2,$C00004
	move.w #$9400, D2
	lsr.l  #8,D1
	move.b D1,D2
	move.w D2,$C00004
	move.w #$9780,$C00004
	move.l #$40000082,(A0)	;DMA fill at VRAM address 0x8000
	nop ;If we uncomment this block, we get corrupted write data at the start of the
	nop ;fill. If we don't have any wait cycles between the control port write and the
	nop ;data port write, the written data is usually correct, with the exception that
	nop ;the byte order of the first write is reversed. If we do insert wait states
	nop ;between the control port write and the data port write, without any wait states
	nop ;before the control port write, the fill starts with a fill data of 0x00, and at
	nop ;some point during the fill operation, usually 4 or 5 bytes in, the data port
	nop ;write then occurs, and the fill operation from then on picks up the correct fill
	nop ;data from that data port write. In other words, if we don't insert a wait state
	nop ;between when we set all our registers for a fill above, and when we actually
	nop ;perform the control port write to setup for the fill operation, the fill can
	nop ;inadvertantly trigger immediately when the control port write is made, without
	nop ;waiting for the data port write to occur. We have found from testing that at
	nop ;least 20 NOP cycles are required in order to get reliable results, when the
	nop ;control port write occurs as a long word write from RAM.
	nop
	nop
	nop
	nop
	nop
	move.l (A0),$C00004
	move.w #$1234,$C00000	;Start DMA fill with 0x1234
;##TODO## The following test for the DMA flag fails on the real hardware. Do more testing
;to determine whether the actual btst operation causes some kind of lock up, or whether
;the result of the operation is simply always 1, and determine the cause. This is
;something we need to emulate, since it has a huge impact.
;##NOTE## We have since realized that we were actually writing 0x8F02 to the control port
;immediately after writing the fill data to the data port, so this may have been the
;cause of our lockups. Do some tests and confirm.
;	DMAWaitLoop1:
;	btst.b #1,$C00005
;	bne DMAWaitLoop1
	DMAWaitLoop1:
	move.w $C00004,D5
	btst #1,D5
	bne DMAWaitLoop1
	move.w #$8144,$C00004
	move.w #$8F02,$C00004

	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000002,$C00004 ;Read from VRAM at 0x8000
	move.w	$C00000,(A0)+	;##NOTE## Has been seen to read back 0x3412!
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+	;##NOTE## Has been seen to read back as 0x0000, when read 9 read as 0x1234!
	move.w	$C00000,(A0)+
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop

	rts
