;-----------------------------------------------------------------------------------------
; Test - FIFO Buffer Size
;-Description:
; This test is designed to demonstrate that the VDP has a 4-word FIFO buffer, and that
; when performing a read from CRAM or VSRAM, the invalid bits from the CRAM and VSRAM
; buffers are populated using the existing data in the next available entry in the FIFO
; buffer.
;-----------------------------------------------------------------------------------------
TestFIFOBufferSizeText:
	dc.b	'FIFO Buffer Size                    '
TestFIFOBufferSizeTextEnd:
TestFIFOBufferSizeExpectedResult:
	dc.w $3000, $3111, $4000, $4000, $5000, $5111, $6000, $6000, $F800, $F111, $F800, $F111, $F800, $F111, $F800, $F111
TestFIFOBufferSize:
	;Initialize the text result block
	move.w	#(TestFIFOBufferSizeTextEnd-TestFIFOBufferSizeText),D7
	move.w	#2,(A0)+ ;Set the number of display lines
	move.w	D7,(A0)+ ;Set the text length
	move.w	#$0020,D6 ;Set the data length
	move.w	D6,(A0)+
	lea		TestFIFOBufferSizeText,A1 ;Write the text to the buffer
	sub.w	#1,D7
	TestFIFOBufferSizeWriteText:
	move.b	(A1)+,(A0)+
	dbf	D7,TestFIFOBufferSizeWriteText
	move.w	#1,(A0)+ ;Set the expected result data present flag
	lea		TestFIFOBufferSizeExpectedResult,A1 ;Write the expected result data to the buffer
	sub.w	#1,D6
	TestFIFOBufferSizeWriteExpectedResultData:
	move.b	(A1)+,(A0)+
	dbf	D6,TestFIFOBufferSizeWriteExpectedResultData

	;Fill the FIFO buffer with some test data by doing some writes to VRAM at 0x8000
	move.l	#$40000002,$C00004
	move.w	#$1111,$C00000
	move.w	#$2222,$C00000
	move.w	#$3333,$C00000
	move.w	#$4444,$C00000
	move.w	#$5555,$C00000
	move.w	#$6666,$C00000

	;Read our data from VSRAM and CRAM through the FIFO into RAM
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$C0200000,$C00004 ;Advance the FIFO buffer
	move.w	#$FFFF,$C00000
	move.l	#$00000010,$C00004 ;Read from VSRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop
	move.l	#$00000020,$C00004 ;Read from CRAM
	move.w	$C00000,(A0)+
	nop ;Wait for a few cycles after a read before modifying any VDP settings. On the
	nop ;real hardware, if a control port write occurs immediately after a data port
	nop ;read, the VDP can lock up. Most likely, this occurs due to the read target
	nop ;being modified to an invalid target before a read pre-cache is complete.
	nop
	nop
	nop
	nop
	nop
	nop

	rts
